#include "as31glue.h"
#include "printf.h"
#include "paulmon2.h"
#include "dirlist.h"
#include "malloc.h"
#include "playlist.h"
#include "display.h"
#include <string.h>
#include "stricmp.h"
#include <8051.h>


// #define VERBOSE_DEBUG


simm_id /* dirlist_t * */ root; // The root dirent_struct
static xdata struct dirent_struct dl_fileinfo;
static dirlist_t * dl_parent;

static dirlist_t * data dl_struct;
static filelist_t * data fl_struct; 
data simm_id temp_id;

// Limit Directory Depth (See dirlist.h for max depth define)
static xdata unsigned char subdir_depth;

/* static*/ xdata unsigned long dirlist_time_count;
/* static*/ xdata unsigned int mp3_file_count;
/* static*/ xdata unsigned int m3u_file_count;
/* static*/ xdata unsigned int total_file_count;
/* static*/ xdata unsigned int directory_count;

static simm_id m3u_list_start=0;
static simm_id m3u_list_end=0;

static xdata char pathname_buffer[320];

#pragma CALLEE-SAVES test_file_extension;
char test_file_extension(void);

void dirlist_read_dir(dirlist_t *this_dir);
void filelist_show(simm_id /* dirlist_t */ dir);


#pragma SAVE
#pragma NOGCSE


void dirlist_add_dir(simm_id /* dirlist_t * */ directory) reentrant
{
  simm_id sub_directory;
  // DO NOT add any non-static local variables here, as they
  // will rapidly overflow the 8051's limited stack space!

  if(++subdir_depth > MAX_SUBDIR_DEPTH) {
        print("Maximum subdir depth reached\r\n");
	subdir_depth--;
	return;
  }

  dirlist_read_dir(addr5(directory));
  //filelist_show(directory);
  sub_directory = ((dirlist_t *)addr5(directory))->subdir_start;

  while(sub_directory != 0) {

      // playlist names are full pathname
      full_pathname(sub_directory);
      temp_id = simm_malloc(strlen(pathname_buffer) + 1);
      strcpy(addr7(temp_id), pathname_buffer);
      playlist_create(temp_id);
      dirlist_add_dir(sub_directory);
      sub_directory = ((dirlist_t *)addr5(sub_directory))->next;
    }
  subdir_depth--;

  // on return from this function the next directory in the queue is processed
}











/* read a directory from the disk drive into a dirlist_struct */
/* and populate all the dirlist_struct and filelist_struct's */
/* that it points to.  The "this_dir" will be passed using */
/* addr5, so this function must not contain any calls to */
/* the addr5 function! */

static xdata simm_id fl_struct_id_state;


/* this bit is set by file_cache_work when it is waiting */
/* for a FAT sector and can not continue to process the file */
/* until that sector is transfered.  The bit is cleared by the */
/* IDE driver every time a sector group is transfered.  This */
/* ugly hack allows us to avoid calling file_cache_work when */
/* it is useless to do so. */
static bit at 0x14 fcw_is_waiting;

static bit fl_struct_already_allocated;
static bit caching_dir;
static bit seek_needed;

void dirlist_read_dir(dirlist_t *this_dir)
{
  xdata char r;
  xdata char dir_fd;
  //xdata char *bu;
  //data simm_id temp_dl;
  xdata char *filename;

#ifdef VERBOSE_DEBUG
  r = SP;
  print("Entered dirlist_read_dir, dir ");
  print_str(addr6(this_dir->long_name));
  print("(depth=");
  print_uint8(subdir_depth);
  print("sp=");
  print_uint8(r);
  print(")\r\n");
#endif

  fl_struct_already_allocated = 0;
  directory_count++;
  dir_fd = file_open_by_1st_cluster(this_dir->cluster);
  file_cache(dir_fd, 0, 0xFFFFFFFF);  // request entire directory cached
  caching_dir = 1;
  seek_needed = 1;

#if 1
  do {
    r = file_cache_work(dir_fd);
    if (r == 2) pm2_exit();
    do_ide_loop(4);
  } while (r == 0);
  do_ide_loop(8);
  file_seek(dir_fd, 0);
#endif

  while(1) {

#if 0
    /* this crazy idea is to be reading the directory and parsing */
    /* it at the same time.  It does not work yet... :(  */
    if (caching_dir && !fcw_is_waiting) {
	r = file_cache_work(dir_fd);
	if (r == 2) pm2_exit();
	if (r) caching_dir = 0;
    }
    if (seek_needed) {
	r = file_seek(dir_fd, 0);
	if (r == 0) {
		seek_needed = 0;
	} else {
		continue;
	}
    }
#endif

    if (fl_struct_already_allocated) {
        temp_id = fl_struct_id_state;
        fl_struct_already_allocated = 0;
    } else {
	temp_id = simm_malloc(sizeof(filelist_t));
    }
    fl_struct = addr7(temp_id); // common case: files rather than directories


    r = dir_read_fast(dir_fd, fl_struct);
    if (r == 1) {
      // print("dir_read_fast returned 1 (xfer pending)\r\n");
      do_ide_loop(4); // data wasn't in cache yet.
      continue;
    }
    if (r) break;     // all other errors, stop reading this directory


#ifdef VERBOSE_DEBUG
    print("attrib=");
    print_hex16(fl_struct->attrib);
    print(":  ");
    /* printf("a=%x, len=%d, ",
         fl_struct->attrib, fl_struct->name_length); */
    /* printf("ext=%c%c%c, ", (char)fl_struct->extension[0],
	   (char)fl_struct->extension[1],
	   (char)fl_struct->extension[2]); */
    /* printf("clust=%lx, ", fl_struct->cluster); */
    /* printf("size=%lx: ", fl_struct->size); */
#endif

    if (fl_struct->attrib == 0x17) {
        print("Skipping RECYCLED directory\r\n");
        continue;
    }
    if ((fl_struct->attrib & 0x0C) != 0) {
        print("Skipping system/hidden file\r\n");
        continue;
    }

    // allocate memory and read filename.  Initial length is rounded
    // up to nearest entry boundary (13) - length is adjusted upon
    // reading of actual entry.
    fl_struct->long_name = simm_malloc(fl_struct->name_length);
    filename = addr6(fl_struct->long_name);
    //fl_struct->name_length = dir_read_fast_name(filename);
    dir_read_fast_name(filename);

#ifdef VERBOSE_DEBUG
    print_str(filename);
    print_crlf();
#endif


    if(fl_struct->attrib & 0x10) {
        if (*filename != '.') {
	  // abnormal case - copy data from fl_struct to dl_struct;
#ifdef VERBOSE_DEBUG
	  print("Processing directory: ");
          print_str(filename);
          print_crlf();
#endif
          fl_struct_id_state = temp_id;  // save id of allocated fl_struct
          fl_struct_already_allocated = 1;

          temp_id = simm_malloc(sizeof(dirlist_t));
          dl_struct = addr6(temp_id);

          // TODO: replace with a fast memcpy
	  memcpy(dl_struct, fl_struct, 17); // copy first 6 items between structs

	  if(this_dir->subdir_start == 0) {
	    this_dir->subdir_start = temp_id;
	    this_dir->subdir_end = temp_id;
	    dl_struct = addr6(temp_id);
	    dl_struct->prev = 0;
	  } else {
	    dl_struct = addr6(this_dir->subdir_end);
	    dl_struct->next = temp_id;
	    dl_struct = addr6(temp_id);
	    dl_struct->prev = this_dir->subdir_end;
	    this_dir->subdir_end = temp_id;
	  }
	  dl_struct->next = 0;
	  dl_struct->subdir_start = 0;
	  dl_struct->subdir_end = 0;
	  dl_struct->filelist_start = 0;
	  dl_struct->filelist_end = 0;
          dl_struct->parent_dir = simm_id_from_pointer(this_dir);
	  dl_struct->open = 1;
	}
    } else {
	total_file_count++;
        switch (test_file_extension()) {
          case 1:  /* MP3 extension */
	    if(this_dir->filelist_start == 0) {
	      this_dir->filelist_start = temp_id;
	      this_dir->filelist_end = temp_id;
              fl_struct->prev = 0;
            } else {
              fl_struct = addr7(this_dir->filelist_end);
              fl_struct->next = temp_id;
              fl_struct = addr7(temp_id);
              fl_struct->prev = this_dir->filelist_end;
              this_dir->filelist_end = temp_id;
	    }
	    fl_struct->next = 0;
	    playlist_add_file(temp_id);   // TODO, move this someday
            mp3_file_count++;
            break;

          case 2:  /* M3U extension */
	    // add m3u file to list to process later
	    if(m3u_list_start == 0) {
	      m3u_list_start = temp_id;
	      m3u_list_end = temp_id;
	      fl_struct->prev=0;
            } else {
              fl_struct = addr7(m3u_list_end);
              fl_struct->next = temp_id;
              fl_struct = addr7(temp_id);
              fl_struct->prev = m3u_list_end;
              m3u_list_end = temp_id;
	    }
	    fl_struct->next=0;
	    print("Detected .m3u playlist file (");
            print_str(filename);
            print(") and added to queue\r\n");
            m3u_file_count++;
            break;

	  default:
	    print("Unknown file extension, ignoring: ");
            print_str(filename);
            print_crlf();
        }
    }
  }
  file_close(dir_fd); 
  free_fat_memory();
}





void m3u_list_show() {
   temp_id = m3u_list_start;

   print("M3U files found:\r\n");
   while (temp_id != 0) {
     fl_struct = addr6(temp_id);
     print_str(addr7(fl_struct->long_name));
     print("cluster=");
     print_hex32(fl_struct->cluster);
     print("size=");
     print_hex32(fl_struct->size);
     printf("ext=%c%c%c\r\n", (char) fl_struct->extension[0],
	    (char) fl_struct->extension[1], (char) fl_struct->extension[2]);
     temp_id = fl_struct->next;
   }
}







// allow for leading /, dirs, and filename parts 
#define MAX_M3U_PATH_PARTS (MAX_SUBDIR_DEPTH+2)
static volatile xdata char * data m3u_ptr;
static unsigned long m3u_size;
static xdata char * m3u_pathname[MAX_M3U_PATH_PARTS];
static unsigned char num_path_parts;

volatile static xdata at 0xFF08 unsigned int block_4000_mapping;
volatile static xdata at 0xFF0A unsigned int block_5000_mapping;
//volatile static xdata at 0xFF0C unsigned int block_6000_mapping;
//volatile static xdata at 0xFF0E unsigned int block_7000_mapping;

unsigned char parse_next_m3u_line(xdata char fd)
{
  if(m3u_ptr > (xdata char *)0x4FFF) {
    block_4000_mapping = block_5000_mapping; // remap 0x5000 block to 0x4000
    m3u_size-=4096;
    if(m3u_size>4096) // load new 0x5000 block in if necessary
      file_mem_map(fd, 5); // map next 4k of file to 'address block' 5
    m3u_ptr-=4096; // change m3u_ptr to point to same position within 0x4000
  }

  num_path_parts = 0;
  m3u_pathname[0] = m3u_ptr;

  while (m3u_ptr < (xdata char *)(0x4000 + m3u_size) /*0x6FFF*/) {
    switch (*m3u_ptr) {
      case '\\':
      case '/':
	//if(start_of_line) // ignore leading / if present
	//  break;
	// hmmm, are .m3u files specified relative to the directory
	// in which they are in? if so, we need to take account of
	// leading slashes to indicate start searching from root dir
	// and store a ptr to the 'current' directory (where the .m3u
	// file is stored as a baseline
	// TODO: store a ptr to the base directory, then if leading /
	//       go to root (in m3u_add_file)
	*m3u_ptr = 0;
	if(++num_path_parts<MAX_M3U_PATH_PARTS) {
	  m3u_pathname[num_path_parts]=m3u_ptr+1;
	} else {
	  print("Too many parts in m3u path name\r\n");
	  // return num_path_parts (not -1) so accesible data is returned,
	  // even if of no use, but also so it still tries next line
	  return num_path_parts;
	}
	break;

      case 10:
      case 13:
	*m3u_ptr = 0;
	m3u_ptr++;
	if (*m3u_ptr == 10 || *m3u_ptr == 13) {
	  *m3u_ptr = 0;
	  m3u_ptr++;
	}
	return (num_path_parts + 1);

      default:
	break;
    }
    m3u_ptr++;
  }
  *m3u_ptr = 0;
  return 0;	// hit end of file...
}





void m3u_add_file() {
  // TODO: search through directories to add this to the playlist
  // TODO: implement linear search just to 'get it going'
  // TODO: work out faster search algorithm (tree/hash table?)
  xdata unsigned char i, r;
  simm_id temp_id2;
  // xdata char *temp_string;

  i=0;
  // TODO: take account of .m3u file's base directory
  //       assume leading / for now
  temp_id=root;
  // TODO: check for leading slash (m3u_pathname[0]=="")
  //       once we start from .m3u file's base dir

  while(temp_id != 0) {
    dl_struct=addr7(temp_id);

    if(i==num_path_parts) {
      print("found correct directory, searching for file ");
      print_str(m3u_pathname[i]);
      print_crlf();

      // traverse files list until match found
      temp_id2 = dl_struct->filelist_start;
      while(temp_id2!=0) {
	fl_struct=addr6(temp_id2);
	/*temp_string=addr7(fl_struct->long_name);
	if((r = mystrcmp(temp_string,m3u_pathname[i])) != 0) {
	  break;
	}*/
      }

      r=0; // temp botch to avoid aslink warning
      // if file found, add file to playlist
      if(r==0) {
	print("found ");
	print_str(m3u_pathname[i]);
	print_crlf();
	//playlist_add_file(/*filelist_struct*/ temp_id);
      } else {
	print("cannot find ");
	print_str(m3u_pathname[i]);
	print_crlf();
	break;
      }
    } else {
      print("searching for ");
      print_str(m3u_pathname[i]);
      print("directory");
      print_crlf();

      // traverse subdirs list until match found
      temp_id2 = dl_struct->subdir_start;
      while(temp_id2!=0) {
	dl_struct=addr6(temp_id2);
	print("comparing ");
	print_str(addr7(dl_struct->long_name));
	print_crlf();
	// strcmp blows DSEG
	// temp_string=addr7(dl_struct->long_name);
	// if((r = stricmp(temp_string,m3u_pathname[i])) != 0)
	//   break;
      }

      r=0; // temp botch to avoid aslink warning
      // if directory not found, give up on this line
      if(r!=0) {
	print("cannot find ");
	print_str(m3u_pathname[i]);
	print("directory");
	print_crlf();
	break;
      }
    }
    i++;
  }
}





void process_m3u() {
  xdata char r, fd, i;
  simm_id head_id;
  simm_id tail_id;
  simm_id file_id;

  file_id=m3u_list_start;
  head_id=0;
  tail_id=0;

#if 0
  // temporary hack to avoid my .m3u test file 1 and go straight to test file 2
  fl_struct=addr7(file_id);
  file_id = fl_struct->next;
#endif

  print("processing m3u files:\r\n");
  while (file_id != 0) {
    fl_struct = addr7(file_id);
    print_str(addr6(fl_struct->long_name));
    print("cluster=");
    print_hex32(fl_struct->cluster);
    print("size=");
    print_hex32(fl_struct->size);
    print_crlf();

    // load m3u file in
    fd = file_open_by_1st_cluster(fl_struct->cluster);
    file_cache(fd, 0, 0xFFFFFFFF);  // request entire file cached
    do {
      do_ide_loop(4);
      r = file_cache_work(fd);
      if (r == 2) pm2_exit();
    } while (r == 0);
    do_ide_loop(8);
    file_seek(fd, 0);

    m3u_size = fl_struct->size;
    // read first 4k of file into 
    file_mem_map(fd, 4); // map first 4k in file to 'address block' 4
    if(m3u_size>4096) {
      // load next 4 block into 0x5000
      file_mem_map(fd, 5);
    }

    // we could make it a condition of processing a .m3u file that it starts
    // with #EXTM3U. I've not seen any without it and there's no other easy
    // way to check its definately a m3u format file (list of directories,
    // comment lines begin with #)
    m3u_ptr = (xdata char *)(0x4000);
    printf("m3u file: [%c,%c,%c]\r\n", (char) m3u_ptr[0], (char) m3u_ptr[1],
      (char) m3u_ptr[2]);

    // create playlist
    print("creating playlist\r\n");
    // TODO: maybe remove .m3u extension from name
    playlist_create(fl_struct->long_name);
    print("created playlist, loading m3u file in\r\n");

    while(parse_next_m3u_line(fd)) {
      if(m3u_pathname[0][0]=='#')
	continue;

      //print_uint8(num_path_parts+1);
      //print("parts: ");
      for(i=0; i<num_path_parts+1; i++) {
	print("[/] ");
	print_str(m3u_pathname[i]);
      }
      print_crlf();
      m3u_add_file(); // TODO: pass directory in which .m3u file is located
    }

    // set up for next m3u file
    file_close(fd);
    // TODO: remove the dirlist entry from the list, and free that memory.
    fl_struct=addr7(file_id); // playlist_add_file uses addr7
    file_id = fl_struct->next;
  }

  print("leaving process_m3u\r\n");
}
























void dirlist_init()
{
  // simm_id temp_id;
  //xdata struct dirstack_struct * this_dirstack_entr;
  unsigned int ticks;

    // debug = 1;

  m3u_list_start=0;
  m3u_list_end=0;
  mp3_file_count=0;
  m3u_file_count=0;
  total_file_count=0;
  directory_count=0;
  dirlist_time_count=clock_tick();

  temp_id = simm_malloc((unsigned int)(sizeof(char) * 2));
  strcpy(addr6(temp_id),"/");

  printf("root long name initalised to %s\r\n", addr6(temp_id));

  subdir_depth = 0;
  root = simm_malloc((unsigned int)(sizeof(struct dirlist_struct)));
  dl_struct = addr6(root);
  dl_struct->subdir_start = 0;
  dl_struct->subdir_end = 0;
  dl_struct->filelist_start = 0;
  dl_struct->filelist_end = 0;
  dl_struct->cluster = get_root_1st_cluster();
  dl_struct->next=0;
  dl_struct->prev=0;
  dl_struct->parent_dir=0;
  dl_struct->long_name=temp_id;
  playlist_create(dl_struct->long_name); // *** reinserted these lines to fix my cock-up 
  dirlist_add_dir(root);    // ***

  ticks = (unsigned int)(clock_tick() - dirlist_time_count);
  printf("\r\n*** Found %u files (%u MP3 & %u M3U) in %u directories",
	total_file_count, mp3_file_count, m3u_file_count, directory_count);
  printf(" in %u.%u seconds ***\r\n\r\n", ticks / 10, ticks % 10);

    // debug = 0;
}







void filelist_show(simm_id /* dirlist_t */ dir)
{
   temp_id = ((dirlist_t *)addr6(dir))->filelist_start;

   while (temp_id != 0) {
	fl_struct = addr6(temp_id);
	printf("%s, %lx\r\n", addr7(fl_struct->long_name), fl_struct->cluster);
	temp_id = fl_struct->next;
   }
}





/* allocate only one copy in xdata, instead of a separate copy */
/* on the (limited) stack for each recursive call */

static xdata unsigned char indent_count;



void dirlist_show_helper(simm_id /* dirlist_t */ dir, unsigned char indent) reentrant
{
  simm_id subdir;

  // files
  dl_parent = addr6(dir);
  temp_id = dl_parent->filelist_start;
  while(temp_id != 0)
    {
      for(indent_count=0; indent_count<indent; indent_count++) print(" ");
      fl_struct = addr6(temp_id);
      print_str(addr7(fl_struct->long_name));
      print_hex32(fl_struct->cluster);
      print_crlf();
      temp_id = fl_struct->next;
    }

  // directories

  subdir = ((dirlist_t *)addr6(dir))->subdir_start;
  while(subdir != 0)
    {
      for(indent_count=0; indent_count<indent; indent_count++) print(" ");
      dl_struct = addr6(subdir);
      print("(dir) ");
      print_str(addr7(dl_struct->long_name));
      print_crlf();
      dirlist_show_helper(subdir, indent+1);
      subdir = ((dirlist_t *)addr6(subdir))->next;
    }
}

void dirlist_show()
{
  unsigned char x = 0;
  print("Dumping directory list...\r\n");

  dirlist_show_helper(root,x);
}






xdata char * full_pathname(simm_id dir)
{
	xdata simm_id dir_chain[MAX_SUBDIR_DEPTH];
	unsigned char num=0;
	xdata char *p, *name;

	/* build a list of all the parent directories */
	do {
		dir_chain[num++] = dir;
		dir = ((dirlist_t *)addr7(dir))->parent_dir;

	} while (dir != 0 && num < MAX_SUBDIR_DEPTH);

	/* traverse the list and build the full pathname string */
	/* TODO: should check for buffer overrun of pathname_buffer */
	p = pathname_buffer;
	do {
		num--;
		if (dir_chain[num] == root) continue;
		*p++ = '/';
		name = addr7(((dirlist_t *)addr6(dir_chain[num]))->long_name);
		while (*name) {
			*p++ = *name++;
		}
	} while (num != 0);
	*p = '\0';

	return pathname_buffer;
}


#pragma RESTORE



// checks the file extension in "fl_struct" (global pointer)
//  return values:
//   0 = no match
//   1 = MP3
//   2 = M3U

char test_file_extension(void)
{
	_asm
	mov	dpl,_fl_struct
	mov	dph,(_fl_struct+1)
	inc	dptr
	inc	dptr
	movx	a, @dptr
	;cjne	a, #'M', test_ext_nope
	cjne	a, #0x4D, test_ext_nope
	inc	dptr
	movx	a, @dptr
	;cjne	a, #'P', test_ext_m3u
	cjne	a, #0x50, test_ext_m3u
	inc	dptr
	movx	a, @dptr
	;cjne	a, #'3', test_ext_nope
	cjne	a, #0x33, test_ext_nope
	mov	dpl, #1
	ret
test_ext_m3u:
	;cjne	a, #'3', test_ext_nope
	cjne	a, #0x33, test_ext_nope
	inc	dptr
	movx	a, @dptr
	;cjne	a, #'U', test_ext_nope
	cjne	a, #0x55, test_ext_nope
	mov	dpl, #2
	ret
test_ext_nope:
	mov	dpl, #0
	ret
	_endasm;
}




