#include "printf.h"
#include "as31glue.h"
#include "parse.h"	// for event codes
#include "string.h"
#include "stricmp.h"
#include "sta013.h"
#include "main.h"
#include "rand.h"

#include "display.h"

#define VOL_TIMEOUT 50
#define SCREEN_TIMEOUT 80

typedef enum {
   MAINSCREEN,
   AUDIOSCREEN,
   EPARAMSCREEN,
   SCREEN_MAX = 2
} screen_t;

typedef enum {
   CONTROL0,
   CONTROL1,
   CONTROL2,
   CONTROL3,
   CONTROL4,
   CONTROL5,
   CONTROL6,
   CONTROL7,
   CONTROL8,
   CONTROL9,
   CONTROL_MAX = 9
} control_t;

static code unsigned char play_mode_description[MODE_MAX+1][11] = {
   " Dir Only ",
   " Dir Rand ",
   " Seqntial ",
   " Plst Rand",
   " All Rand "};

#define DISPLAY_WIDTH 24
#define FILEBUF_MAX 256
#define PLAYBUF_MAX 256

simm_id current_filename;
simm_id current_playlistname;

xdata unsigned char fxwrap;		// Position inside filebuf
xdata unsigned char fxscroll;	// Scroll ?
xdata unsigned char filebuf[FILEBUF_MAX];
xdata unsigned char fdisplaybuf[DISPLAY_WIDTH + 1];
xdata unsigned char filebuf_len;

xdata unsigned char pxwrap;               // Position inside playbuf
static bit pxscroll;	        // Scroll ?
xdata unsigned char playbuf[PLAYBUF_MAX];
xdata unsigned char pdisplaybuf[DISPLAY_WIDTH + 1];
xdata unsigned char playbuf_len;

xdata struct id3_tag_struct id3_tag;

// TODO: Remove this once display bug is fixed
xdata update_counter;

screen_t current_screen;
control_t current_control;

// display version
xdata unsigned char lcd_version;

//--------------------------------------------------------------------
//	lcd_update_2nd_line
//
//	Called when timer 0 expires.
//	Displays fdisplaybuf on the second line of the display, then
//	scrolls fdisplaybuf to the left, and grabs the next
//	character from filebuf.
//
void lcd_update_2nd_line(const unsigned char refresh)
{
   xdata unsigned char font;

   if (current_control == CONTROL3)
      font = 33;  // bold
   else
      font = 32;  // normal
	
   if (lcd_version == 0)
   {
      if (fxscroll == 0)
      {
	 // not scrolling so just display it
	 printf("\\[\\B !\\C%c%s\\C \\]\r\n", (char) font, filebuf);
      }
      else
      {
	 printf("\\[\\B !\\C%c%s\\C \\]\r\n", (char) font, fdisplaybuf);

	 // shift buffer & insert new char
	 fxwrap++;
	 memcpy(fdisplaybuf, fdisplaybuf+1, DISPLAY_WIDTH - 1);

	 if(fxwrap < filebuf_len) {
	    fdisplaybuf[DISPLAY_WIDTH-1] = filebuf[fxwrap];
	 } else if (fxwrap < filebuf_len + 4) {
	    fdisplaybuf[DISPLAY_WIDTH-1] = ' ';
	 } else {
	    fxwrap = 0;
	    fdisplaybuf[DISPLAY_WIDTH-1] = filebuf[0];
	 }
      }

      // TODO: Remove this once display bug is fixed
      if((update_counter++)>10)
      {
	 update_counter=0;
	 lcd_change_menu(0);
      }
   }
   else
   {
      // ensure font is correct upon refresh (selection might have changed)
      if (refresh != 0)
      {
	 if (fxscroll != 0) 
	 {
	    printf("\\K2!1\\K1!%c", (char) font);
	 }
	 else
	 {
	    printf("\\[\\B !\\C%c%s\\C \\]\r\n", (char) font, filebuf);
	 }
      }
   }
}

//--------------------------------------------------------------------
//
//	lcd_display_filename
//
//	Takes long file name.  Strips last 4 characters if
//	they are ".m3u".  Pads with a few spaces and
//	fills to DISPLAY_WIDTH characters if shorter than that.
//
void lcd_display_filename()
{
   xdata unsigned char* source;
   xdata unsigned char* dest;

   if (current_filename == 0)
   {
      return;
   }

   // make local copy (filename will be invalid on exit)
   // check length & truncate.  convert '_' to ' '.
   source = addr7(current_filename);
   dest = filebuf;
   filebuf_len = 0;
   while(*source != '\0')
   {
      filebuf_len++;
      if (filebuf_len >= FILEBUF_MAX)
      {
	 break;
      }
      else if (*source == '_')
      {
	 *dest = ' ';
      }
      else
      {
	 *dest = *source;
      }

      source++;
      dest++;
   }
   *dest = '\0';

   // .mp3 and .MP3
   if (filebuf_len > 4 && str_is_dot_mp3(filebuf + filebuf_len - 4))
   {
      filebuf_len -= 4;
      *(filebuf + filebuf_len) = '\0';
   }
	
   if (filebuf_len <= DISPLAY_WIDTH)
   {
      // pad with maximum number of spaces
      memset(filebuf + filebuf_len, ' ', DISPLAY_WIDTH);
      filebuf[DISPLAY_WIDTH] = '\0';
      fxscroll = 0;
   }
   else
   {
      // init scroll buffer
      memcpy(fdisplaybuf, filebuf, DISPLAY_WIDTH);
      fdisplaybuf[DISPLAY_WIDTH] = '\0';
      fxscroll = 1;
   }

   if (lcd_version == 0)
   {
      fxwrap = 23;
      lcd_update_2nd_line(0);
   }
   else
   {
      if (fxscroll == 0)
      {
	 // scrolling off and force display of new string
	 print("\\K2!0");
	 lcd_update_2nd_line(1);
      }
      else
      {
	 xdata unsigned char font;

	 if (current_control == CONTROL3)
	    font = 33;  // bold
	 else
	    font = 32;  // normal
	
	 // just fill scrolling buffer and let the display do the work
	 printf("\\[\\K1!%c\\K3!1%s    \\K3!0\\K2!1\\]", (char) font, filebuf);
      }
   }
}


//--------------------------------------------------------------------
void lcd_set_filename(simm_id name)
{
   current_filename = name;
   if (current_screen == MAINSCREEN || current_screen == AUDIOSCREEN)
   {
      lcd_display_filename();
   }
}

//--------------------------------------------------------------------
//
//	lcd_update_3rd_line
//
//	Called when timer 0 expires.
//	Displays pdisplaybuf on the third line of the display, then
//	scrolls pdisplaybuf to the left, and grabs the next
//	character from playbuf.
//
void lcd_update_3rd_line(const unsigned char refresh)
{
   xdata unsigned char font;

   if (current_control == CONTROL4)
      font = 33;  // bold
   else
      font = 32;  // normal
	
   if (lcd_version == 0)
   {
      if (pxscroll == 0)
      {
	 // not scrolling so just display it
	 printf("\\[\\B \"\\C%c%s\\C \\]\r\n", (char) font, playbuf);
      }
      else
      {
	 printf("\\[\\B \"\\C%c%s\\C \\]\r\n", (char) font, pdisplaybuf);

	 // shift buffer & insert new char
	 pxwrap++;
	 memcpy(pdisplaybuf, pdisplaybuf+1, DISPLAY_WIDTH - 1);

	 if(pxwrap < playbuf_len) {
	    pdisplaybuf[DISPLAY_WIDTH-1] = playbuf[pxwrap];
	 } else if (pxwrap < playbuf_len + 4) {
	    pdisplaybuf[DISPLAY_WIDTH-1] = ' ';
	 } else {
	    pxwrap = 0;
	    pdisplaybuf[DISPLAY_WIDTH-1] = playbuf[0];
	 }
      }
   }
   else
   {
      // ensure font is correct upon refresh (selection might have changed)
      if (refresh != 0)
      {
	 if (pxscroll != 0) 
	 {
	    printf("\\K2\"1\\K1\"%c", (char) font);
	 }
	 else
	 {
	    printf("\\[\\B \"\\C%c%s\\C \\]\r\n", (char) font, playbuf);
	 }
      }
   }
}

//--------------------------------------------------------------------
//
//	lcd_display_playlist_name
//
//	Takes long file name.  Strips last 4 characters if
//	they are ".m3u".  Pads with a few spaces and
//	fills to DISPLAY_WIDTH characters if shorter than that.
//
void lcd_display_playlistname()
{

   xdata unsigned char* source;
   xdata unsigned char* dest;

   if (current_playlistname == 0)
   {
      return;
   }

   // make local copy (pl_name will be invalid on exit)
   // check length & truncate.  convert '_' to ' '.
   source = addr7(current_playlistname);
   dest = playbuf;
   playbuf_len = 0;
   while(*source != '\0')
   {
      playbuf_len++;
      if (playbuf_len >= PLAYBUF_MAX)
      {
	 break;
      }
      else if (*source == '_')
      {
	 *dest = ' ';
      }
      else
      {
	 *dest = *source;
      }

      source++;
      dest++;
   }
   *dest = '\0';

   // .m3u and .M3U
   if (playbuf_len > 4 && str_is_dot_m3u(playbuf + playbuf_len - 4))
   {
      playbuf_len -= 4;
      *(playbuf + playbuf_len) = '\0';
   }
	
   if (playbuf_len <= DISPLAY_WIDTH)
   {
      // pad with maximum number of spaces
      memset(playbuf + playbuf_len, ' ', DISPLAY_WIDTH);
      playbuf[DISPLAY_WIDTH] = '\0';
      pxscroll = 0;
   }
   else
   {
      // init scroll buffer
      memcpy(pdisplaybuf, playbuf, DISPLAY_WIDTH);
      pdisplaybuf[DISPLAY_WIDTH] = '\0';
      pxscroll = 1;
   }

   if (lcd_version == 0)
   {
      pxwrap = 23;
      lcd_update_3rd_line(0);
   }
   else
   {
      if (pxscroll == 0)
      {
	 // scrolling off and force display of new string
	 print("\\K2\"0");
	 lcd_update_3rd_line(1);
      }
      else
      {
	 xdata unsigned char font;

	 if (current_control == CONTROL4)
	    font = 33;  // bold
	 else
	    font = 32;  // normal
	
	 // just fill scrolling buffer and let the display do the work
	 printf("\\[\\K1\"%c\\K3\"1%s    \\K3\"0\\K2\"1\\]", (char) font, playbuf);
      }
   }
}

//--------------------------------------------------------------------
void lcd_set_playlistname(simm_id name)
{
   current_playlistname = name;
   if (current_screen == MAINSCREEN || current_screen == AUDIOSCREEN)
   {
      lcd_display_playlistname();
   }
}


//--------------------------------------------------------------------
// called on expiry of timer 2
void lcd_timer_callback()
{
   if (lcd_version == 0)
   {
      if (current_screen == MAINSCREEN || current_screen == AUDIOSCREEN)
      {
	 lcd_update_2nd_line(0);
	 lcd_update_3rd_line(0);
      }
   }
}

//--------------------------------------------------------------------
void lcd_display_id3_struct() 
{

   id3_tag.title[DISPLAY_WIDTH] = 0;
   id3_tag.artist[DISPLAY_WIDTH] = 0;
   id3_tag.album[DISPLAY_WIDTH] = 0;


   // Display ID3 data
   // use display clear line functions with version 1 displays
   if (current_screen == MAINSCREEN)
   {
      printf("\\L#\\L$\\L%%\\[\\B #%s\\B $%s\\B %%%s\\]\r\n", 
	     id3_tag.title, 
	     id3_tag.artist, 
	     id3_tag.album);
   }
}

//--------------------------------------------------------------------
void lcd_display_id3_init()
{
   // reset id3 data at the start of each new track and when no id3 data present
   id3_tag.title[0] = 0;
   id3_tag.artist[0] = 0;
   id3_tag.album[0] = 0;

   // force display update
   lcd_display_id3_struct();
}

//--------------------------------------------------------------------
void lcd_update_time() {
	xdata unsigned char s[6];
	xdata unsigned char min;
	xdata unsigned int sec;	
	static xdata unsigned int prevsec = 0xffffffff;

	// Now get the elapsed time to display.  Subtract the number of frames played when
	// this song started (updated in main.c when the song changes) from the current number
	// of frames played back.  Multiply that by the number of milliseconds per from for the 
	// current sample rate (defined in the playback struct in display.c).  Then divide the whole
	// mess by 1000 to get seconds.  Break it up to seconds & minutes, then output.
	// -ZSB 12-Aug-2001 3:33 PM
		
	sec = ((GetFrameCount() - songStartFrameCount) * playback.ms_per_frame) / 1000;
#if 0
	if (sec != prevsec)
#endif
	{
	   prevsec = sec;

	   min = sec / 60;
	   sec = sec % 60;

	   if(min < 10) {
	      s[0] = '0';
	   } else {
	      s[0] = (min / 10) + '0';
	   }
	
	   s[1] = (min % 10) + '0';

	   s[2] = ':';

	   if(sec < 10) {
	      s[3] = '0';
	   } else {
	      s[3] = (sec / 10) + '0';
	   }
	
	   s[4] = (sec % 10) + '0';

	   s[5] = 0;

	   printf("\\[\\B3 %s\\]\r\n",s);
	}
}






//--------------------------------------------------------------------
//
//	lcd_navigation
//
//	Display navigation element on the screen at coordinates (x,y).
//	If this_control is the active one then display in bold.
//
void lcd_navigation(unsigned char x,
		    unsigned char y,
		    const unsigned char *navtext, 
		    unsigned char this_control)
{
	if(this_control==current_control)
	{
		printf("\\[\\B%c%c\\C!%s\\C \\]\r\n",(char)(x+32),(char)(y+32),navtext);
	}
	else
	{
		printf("\\[\\B%c%c\\C %s\\C \\]\r\n",(char)(x+32),(char)(y+32),navtext);
	}
}




//--------------------------------------------------------------------
void lcd_update_top_line(void) 
{	
   if (current_screen == MAINSCREEN || current_screen == AUDIOSCREEN)
   {
	if (playing) {
		lcd_navigation(0,0,"Playing ",CONTROL1);
	} else {
		lcd_navigation(0,0,"Paused  ",CONTROL1);
	}

	lcd_navigation(8, 0, play_mode_description[play_mode], CONTROL2);
   }
}

//--------------------------------------------------------------------
//
//	lcd_bar_set
//
//	Displays a bargraph on a line of
//	the display.
//
//	The sta013 expects a 2s compliment number from
//	-12 to 12 corresponding to a boost of
//	-18dB to 18dB.
//
void lcd_bar_set(param_t param)
{
	xdata unsigned char bar_position;
	xdata unsigned char display_line;
	xdata unsigned char font;

        switch(param)
	{
	   case PARAM_TREBLE:
	      lcd_navigation(0, 4, "Tr:", CONTROL5);
	      if (current_control == CONTROL5)
		 font = 33;  // bold
	      
	      else
		 font = 32;  // normal
	      display_line = '$';
	      break;
	   case PARAM_BASS:
	      lcd_navigation(0, 5, "Ba:", CONTROL6);
	      if (current_control == CONTROL6)
		 font = 33;  // bold
	      
	      else
		 font = 32;  // normal
	      display_line = '%';
	      break;
	   default:
	      return;
	}

	bar_position = (param_value[param] + STA013_MAX_TONE - STA013_PARAM_CENTRE);

	// adjust to within the 21 chars available (range is -12 <-> +12 ie 25 units)
	if (bar_position < 2)
	{
	   bar_position = 0;
	}
	else if (bar_position > 21)
	{
	   bar_position = 20;
	}
	else
	{
	   bar_position -= 2;
	}
	
	printf("\\[\\C%c\\B#%c----------%c----------\\B%c%c\\C!\\w\\C \\]\r\n",
	       (char) font,
	       (char) display_line,
	       (char) 247,
	       (char) (' ' + 3 + bar_position),
	       (char) display_line);
}




//--------------------------------------------------------------------
//
//	lcd_volume_set_db
//
//	Displays a volume db at the specified coordinates.
//
void lcd_volume_set_db()
{
   xdata unsigned char attn;

   attn = param_value[PARAM_ATTN] - STA013_PARAM_CENTRE + tone_attn;

	if(attn==0)
	{
		print("\\L'\\[\\B%'  0 dB\\]\r\n");
	}
	else
	{
		printf("\\L'\\[\\B%%'-%2d dB\\]\r\n",attn);
	}

	lcd_navigation(0, 7,"Vol: ",CONTROL0);
}


//--------------------------------------------------------------------
//
//	lcd_change_menu
//
//	Clears display and displays the initial data
//	for the current screen.  Increments current screen
//	if action==1;
//
void lcd_change_menu(unsigned char action)
{
	// Some sanity checks
	if(action==1)
	{
		if((++current_screen) > SCREEN_MAX)
		{
		   current_screen=MAINSCREEN;
		   current_control = CONTROL0;
		}
	}

	if (lcd_version != 0)
	{
	   // display updates off
	   print("\\I:");
	}

	// Clear display
	print("\\@ 0");

	switch(current_screen)
	{
	  case MAINSCREEN:
		lcd_update_top_line();
		lcd_display_filename();
		lcd_display_playlistname();
		lcd_screen_dynamic(1);
		break;

	  case AUDIOSCREEN:
		lcd_update_top_line();
		lcd_update_2nd_line(1);
		lcd_update_3rd_line(1);
		lcd_screen_dynamic(1);
		break;

	  case EPARAMSCREEN:
		lcd_screen_dynamic(1);
 		break;

	  default:
		break;
	}

	if (lcd_version != 0)
	{
	   // redraw whole display & display updates on
	   print("\\I8\\I9");
	}
}




//--------------------------------------------------------------------
// update and restart the backlight timer back upon the parameter
// setting

void lcd_update_backlight_timer()
{
   switch(param_value[PARAM_BACKLIGHT_TIMEOUT])
   {
      case 0:
	 // Never
	 print("\\J20\\J00");
	 break;
      case 1:
	 // Always on
	 print("\\J21\\J01");
	 break;
      default:
	 // TODO : need to implement timer
	 break;
   }
}


//--------------------------------------------------------------------
void lcd_setup(void)
{
	print("lcd:");

	lcd_set_scroll_speed();
	parse_event_setup();
	lcd_update_backlight_timer();

	// setup navigation
	current_screen = MAINSCREEN;
	current_control = CONTROL0;

	// ask display for version
	lcd_version = 0;
	print("\\E\r\n");

	string_init(filebuf, ' ', FILEBUF_MAX);
	string_init(fdisplaybuf, ' ', DISPLAY_WIDTH + 1);
	string_init(playbuf, ' ', FILEBUF_MAX);
	string_init(pdisplaybuf, ' ', DISPLAY_WIDTH + 1);

	pxscroll = 0;
	playbuf_len = 0;
	pxwrap = 0;

	fxscroll = 0;
	filebuf_len = 0;
	fxwrap = 0;

	current_filename = 0;
	current_playlistname = 0;
}

//--------------------------------------------------------------------
void lcd_screen_dynamic(const unsigned char refresh)
//
//	lcd_screen_dynamic
//
//	Update dynamic data.
//
{

   static xdata unsigned int last_bitrate = 0;
   static xdata unsigned char last_freq = 0;
   static xdata unsigned char last_mode = 0;

	switch(current_screen)
	{
	  case MAINSCREEN:
		// Now this bit should only update the screen if it "has to."  IE if an item has changed or if
		// we're doing a screen refresh. -ZSB 18-Aug-2001 2:29 PM
		playback.mode_known = 0; // force update_playback_params to reget the info for the new bitrate (if applicable)
		update_playback_parms(); // This should get any new bitrate for VBR's
		// Is there any way to tell if an MP3 is VBR or not?  If so, we could skip calling this for CBR files.

		if (playback.mode_known) {
		        // On a refresh or change, we send out the whole thing
			if(refresh
			   || last_bitrate != playback.bitrate
			   || last_freq != playback.freq
			   || last_mode != playback.mode[0]) {
				last_bitrate = playback.bitrate;
				last_mode = playback.mode[0];
				last_freq = playback.freq;
				printf("\\[\\B &%3d kbps  %2d kHz  %s\\]", last_bitrate, playback.freq, playback.mode);
			}
		} else { 
		        // We'll always display blanks, regardless of the
		        // bitrate settings if things are undefined.
			// Just to make sure we clear the screen quickly....
			print("\\[\\B &    kbps     kHz        \\]");
			last_bitrate = 0;
			last_freq = 0;
			last_mode = 0;
		}

		if (refresh) {
		   lcd_display_id3_struct();
		   lcd_volume_set_db();
		}

		lcd_update_time();
		break;

	  case AUDIOSCREEN:
	        if (refresh)
	        {
		   lcd_bar_set(PARAM_TREBLE);
		   lcd_bar_set(PARAM_BASS);
		   lcd_volume_set_db();
		}
		lcd_update_time();
		break;

	  case EPARAMSCREEN:
                if (refresh)
                {
		   if (lcd_version > 0)
		   {
		      print("\\B  \\[MP3 Player Configuration\\]");
		      eparam_display();
		   }
		   else
		   {
		      print("\\B $\\[  Configuration Screen  \\]");
		      print("\\B %\\[is not supported on beta\\]");
		      print("\\B &\\[        displays        \\]");
		   }
                }
		break;

	  default:
		break;
	}

}



//--------------------------------------------------------------------
void lcd_change(param_t parm)
{
   switch (parm) {
      case PARAM_ATTN:
	 lcd_volume_set_db();
	 break;
	 
      case PARAM_TREBLE:
	 // changes to the treble setting can change the volume
	 lcd_bar_set(PARAM_TREBLE);
	 lcd_volume_set_db();
	 break;

      case PARAM_BASS:
	 // changes to the bass setting can change the volume
	 lcd_bar_set(PARAM_BASS);
	 lcd_volume_set_db();
	 break;
	 
      case PARAM_PLAY_MODE:
	 lcd_update_top_line();
	 break;
	 
      default:
	 printf("Unknown LCD change event: %d\n", parm);
   }
}

//--------------------------------------------------------------------
// Main Screen navigation event handler
//
void user_action_screen1(event_t event)
{
   static code up_control[CONTROL_MAX+1] = {
      CONTROL4, CONTROL0, CONTROL0, CONTROL1,
      CONTROL3, CONTROL5, CONTROL6, CONTROL7,
      CONTROL8, CONTROL9};

   static code down_control[CONTROL_MAX+1] = {
      CONTROL1, CONTROL3, CONTROL3, CONTROL4,
      CONTROL0, CONTROL5, CONTROL6, CONTROL7,
      CONTROL8, CONTROL9};

   switch(event) {
      case E_RIGHT:
	 if (current_control == CONTROL4) {
	    current_control = CONTROL0;
	 } else {
	    current_control++;
	 }
	 lcd_change_menu(0);
	 break;
      case E_LEFT:
	 if (current_control == CONTROL0) {
	    current_control = CONTROL4;
	 } else {
	    current_control--;
	 }
	 lcd_change_menu(0);
	 break;
      case E_UP:
	 current_control = up_control[current_control];
	 lcd_change_menu(0);
	 break;
      case E_DOWN:
	 current_control = down_control[current_control];
	 lcd_change_menu(0);
	 break;
      case E_INC:
	 switch(current_control) {
	    case CONTROL0:		// Volume
	       action_vol_up();
	       break;
	    case CONTROL1:		// Play/Pause
	       put_back_event(E_PLAY_PAUSE);
	       break;
	    case CONTROL2:		// Normal/Random
	       random_mode_next();
	       lcd_update_top_line();
	       break;
	    case CONTROL3:		// Filename
	       put_back_event(E_NEXT);
	       break;
	    case CONTROL4:		// Directory/Playlist
	       put_back_event(E_NEXT_LIST);
	       break;
	    default:
	       break;
	 }
	 break;
      case E_DEC:
	 switch(current_control) {
	    case CONTROL0:		// Volume
	       action_vol_down();
	       break;
	    case CONTROL1:		// Play/Pause
	       put_back_event(E_PLAY_PAUSE);
	       break;
	    case CONTROL2:		// Normal/Random
	       random_mode_prev();
	       lcd_update_top_line();
	       break;
	    case CONTROL3:		// Filename
	       put_back_event(E_PREV);
	       break;
	    case CONTROL4:		// Directory/Playlist
	       put_back_event(E_PREV_LIST);
	       break;
	    default:
	       break;
	 }
	 break;
      default:
	 printf("Unsupported event %d\r\n", event);
	 break;
   }
}

//--------------------------------------------------------------------
// Audio screen navigation event handler
//
void user_action_screen2(event_t event)
{
   static code up_control[CONTROL_MAX+1] = {
      CONTROL6, CONTROL0, CONTROL0, CONTROL1,
      CONTROL3, CONTROL4, CONTROL5, CONTROL6,
      CONTROL7, CONTROL9};

   static code down_control[CONTROL_MAX+1] = {
      CONTROL1, CONTROL3, CONTROL3, CONTROL4,
      CONTROL5, CONTROL6, CONTROL0, CONTROL8,
      CONTROL9, CONTROL0};

   switch(event) {
      case E_RIGHT:
	 if (current_control == CONTROL6) {
	    current_control = CONTROL0;
	 } else {
	    current_control++;
	 }
	 lcd_change_menu(0);
	 break;
      case E_LEFT:
	 if (current_control == CONTROL0) {
	    current_control = CONTROL8;
	 } else {
	    current_control--;
	 }
	 lcd_change_menu(0);
	 break;
      case E_UP:
	 current_control = up_control[current_control];
	 lcd_change_menu(0);
	 break;
      case E_DOWN:
	 current_control = down_control[current_control];
	 lcd_change_menu(0);
	 break;
      case E_INC:
	 switch(current_control) {
	    case CONTROL0:		// Volume
	       action_vol_up();
	       break;
	    case CONTROL1:		// Play/Pause
	       put_back_event(E_PLAY_PAUSE);
	       break;
	    case CONTROL2:		// Random Mode
	       random_mode_next();
	       lcd_update_top_line();
	       break;
	    case CONTROL3:		// Filename
	       put_back_event(E_NEXT);
	       break;
	    case CONTROL4:		// Directory/Playlist
	       put_back_event(E_NEXT_LIST);
	       break;
	    case CONTROL5:		// Treble
	       sta013_tone_up(PARAM_TREBLE);
	       break;
	    case CONTROL6:		// Bass
	       sta013_tone_up(PARAM_BASS);
	       break;
	    default:
	       break;
	 }
	 break;
      case E_DEC:
	 switch(current_control) {
	    case CONTROL0:		// Volume
	       action_vol_down();
	       break;
	    case CONTROL1:		// Play/Pause
	       put_back_event(E_PLAY_PAUSE);
	       break;
	    case CONTROL2:		// Normal/Random
	       random_mode_prev();
	       lcd_update_top_line();
	       break;
	    case CONTROL3:		// Filename
	       put_back_event(E_PREV);
	       break;
	    case CONTROL4:		// Directory/Playlist
	       put_back_event(E_PREV_LIST);
	       break;
	    case CONTROL5:		// Treble
	       sta013_tone_down(PARAM_TREBLE);
	       break;
	    case CONTROL6:		// Bass
	       sta013_tone_down(PARAM_BASS);
	       break;
	    default:
	       current_screen=MAINSCREEN;
	       break;
	 }
	 break;
      default:
	 printf("Unsupported event %d\r\n", event);
	 break;
   }
}

//--------------------------------------------------------------------
void lcd_set_scroll_speed()
{
   // do with updates off to reduce flicker
   printf("\\I:\\K0%d\\I8\\I9", (int) param_value[PARAM_SCROLL_SPEED]);
}
   
//--------------------------------------------------------------------
void lcd_user_action(event_t event)
{
   switch(event) {
      case E_DISPLAY1:
	 lcd_version = 1;
	 printf("\r\nDisplay protocol version=%d.\r\n", lcd_version);
	 break;

      case E_ENTER:
	 set_timer(TIMER_LCD_TIMEOUT, param_screen_timeout());
	 lcd_change_menu(1);            // Enter always changes pages
	 break;

      case E_LCD_TIMEOUT_TIMER:
	 if(current_screen!=MAINSCREEN || 
	    current_control!=CONTROL0) {
	    current_screen=MAINSCREEN;
	    current_control=CONTROL0;
	    lcd_change_menu(0);
	 }
	 break;

      default:
	 set_timer(TIMER_LCD_TIMEOUT, param_screen_timeout());
	 // all other events are screen specific and sent to a particular screen
	 switch(current_screen) {
	    case MAINSCREEN:
	       user_action_screen1(event);
	       break;
	    case AUDIOSCREEN:
	       user_action_screen2(event);
	       break;
	    case EPARAMSCREEN:
	       if (eparam_user_action(event))
	       {
		  lcd_change_menu(0);
	       }
	       break;
	    default:
	       current_screen=MAINSCREEN;
	       break;
	 }
	 break;
   }
}

